#include "../include/Server.h"
#include "../include/Config.h"
#include "../include/base64.h"

#include <iostream>

using namespace boost::asio;
using ip::tcp;

/***************
 * con_handler *
 ***************/

con_handler::con_handler(
    basic_socket_acceptor<ip::tcp>::executor_type &io_service)
    : sock(io_service), buf(max_length), jsonCommander(fileManager) {
  // disable indentation for json
  this->jsonStringBuilder.settings_["indentation"] = "";

  // init reader to parse json
  Json::CharReaderBuilder builder;
  jsonReader = std::unique_ptr<Json::CharReader>(builder.newCharReader());
}

con_handler::~con_handler() {}

con_handler::pointer
con_handler::create(basic_socket_acceptor<ip::tcp>::executor_type &io_service) {
  return pointer(new con_handler(io_service));
}

tcp::socket &con_handler::socket() { return sock; }

void con_handler::start() { read(&con_handler::handle_read_version); }

void con_handler::handle_read_version(const boost::system::error_code &err,
                                      size_t bytes_transferred) {
  if (!err) {
    // set up json stuff
    Json::Value root = parseMessage();

    JsonCommander::Response response = this->jsonCommander.testVersion(root);

    switch (response.action) {
    case JsonCommander::Action::send:
      read(&con_handler::handle_read_login);
      sendJson(response.json);
      break;
    case JsonCommander::Action::closeAndSend:
      sendJson(response.json);
    default:
      sock.close();
    }

  } else {
    std::cerr << "error: " << err.message() << std::endl;
    sock.close();
  }
}

void con_handler::handle_read_login(const boost::system::error_code &err,
                                    size_t bytes_transferred) {
  if (!err) {
    // set up json stuff
    Json::Value root = parseMessage();

    JsonCommander::Response response = this->jsonCommander.checkLogin(root);

    switch (response.action) {
    case JsonCommander::Action::send:
      read(&con_handler::handle_read_command);
      sendJson(response.json);
      break;
    case JsonCommander::Action::closeAndSend:
      sendJson(response.json);
    default:
      sock.close();
    }

  } else {
    std::cerr << "error: " << err.message() << std::endl;
    sock.close();
  }
}

void con_handler::handle_read_command(const boost::system::error_code &err,
                                      size_t bytes_transferred) {
  if (!err) {
    // set up json stuff
    Json::Value root = parseMessage();

    JsonCommander::Response response = this->jsonCommander.execute(root);

    switch (response.action) {
    case JsonCommander::Action::send:
      read(&con_handler::handle_read_command);
      sendJson(response.json);
      break;
    case JsonCommander::Action::closeAndSend:
      sendJson(response.json);
    default:
      sock.close();
    }

  } else {
    std::cerr << "error: " << err.message() << std::endl;
    sock.close();
  }
}

void con_handler::handle_write(const boost::system::error_code &err,
                               size_t bytes_transferred) {
  if (!err) {
    std::cout << "Hello World!" << std::endl;
  } else {
    std::cerr << "error: " << err.message() << std::endl;
    sock.close();
  }
}

void con_handler::read(void (con_handler::*handler)(
    const boost::system::error_code &err, size_t bytes_transferred)) {
  /*sock.async_read_some(buffer(data, max_length),
                       boost::bind(handler, shared_from_this(),
                                   placeholders::error,
                                   placeholders::bytes_transferred));*/

  async_read_until(sock, buf, '\n',
                   bind(handler, shared_from_this(), placeholders::error,
                        placeholders::bytes_transferred));
}

void con_handler::sendJson(const Json::Value &json) {
  std::string jsonString = Json::writeString(jsonStringBuilder, json);
  jsonString.append("\n");

  sock.async_write_some(buffer(jsonString, max_length),
                        boost::bind(&con_handler::handle_write,
                                    shared_from_this(), placeholders::error,
                                    placeholders::bytes_transferred));
}

Json::Value con_handler::parseMessage() {
  const char *data = buffer_cast<const char *>(buf.data());

  std::string dataStr(data, buf.size());
  const int lineEnd = dataStr.find('\n');

  JSONCPP_STRING err;
  Json::Value root;

  // parse data
  if (!this->jsonReader->parse(data, data + lineEnd, &root, &err)) {
    std::cerr << "Json error: " << err << std::endl << "data: " << data;
    sock.close();
  }

  buf.consume(lineEnd + 1);

  return root;
}

/**********
 * Server *
 **********/

void Server::start_accept() {
  auto executor = acceptor_.get_executor();
  con_handler::pointer connection = con_handler::create(executor);
  acceptor_.async_accept(connection->socket(),
                         boost::bind(&Server::handle_accept, this, connection,
                                     placeholders::error));
}

Server::Server(io_service &io_service)
    : acceptor_(io_service,
                tcp::endpoint(tcp::v4(), std::stoi(Config::getValue("port")))) {
  start_accept();
}

Server::~Server() {}

void Server::handle_accept(con_handler::pointer connection,
                           const boost::system::error_code &err) {
  if (!err) {
    connection->start();
  }
  start_accept();
}
